/*
Programm: Aufg.AusdrAnalyseBruch
Zweck:    Berechnung des Wertes vollstaendig geklammerter 
          Ausdruecke mit Bruechen
Aufgabe:  Entspricht dem Punkt "8.9.6 Aufgabe 6" im 
          Skript "Programmieren 1" von Prof. Letschert, in der 
          Version vom 2001-09-10
Autor:    Mattias Ansorg <ansis@gmx.de>
Lizenz:   public domain
Datum:    2001-12-10
*/


#include <iostream>
#include<string>
using namespace std;

namespace BruchNS {
   enum Vorzeichen {plus, minus};
   struct Bruch {
     Vorzeichen vz;
     int zaehler;
     int nenner;
     void lese();
     void schreibe();
     void kuerze(); // soweit wie moeglich kuerzen
     void erweitere(int);
     void kehrwert();
  };

   Bruch operator+(Bruch, Bruch); 
   Bruch operator-(Bruch, Bruch); 
   Bruch operator*(Bruch, Bruch); 
   Bruch operator/(Bruch, Bruch); 
   void gleichnamig(Bruch &, Bruch &);
   int kgv(int, int);
   int ggt(int, int);   
} 

using namespace BruchNS;

Bruch ausdruckWert(string, int &);
Bruch bruchWert(string, int &);
void skipspace(string, int &);
bool istZiffer(char);
int zahlWert(string, int &);

int main () {
   string a;
   Bruch wert;
   int pos=0;
   cout << "Bitte Bruchausdruck eingeben:\n";
   getline(cin,a); // Ausdruck soll nicht bei Leerzeichen enden
   wert = ausdruckWert(a,pos);
   cout << "Wert des Ausdrucks:\n";
   wert.kuerze();
   wert.schreibe();
} 

Bruch ausdruckWert(string ausdruck, int &pos) {
   // WARNUNG: Funktion prueft nicht vollstaendig auf Korrektheit der Ausdruecke
   //          bei falschen Ausdruecken ggf. Absturz
   // pos: Zeichen an Position pos in ausdruck
   // pos==ausdruck.length(): StringEnde erreicht
   // rekursive Auswertung nach der Grammatik:
   // <ausdruck>= [<space>] 
   //             (<bruch> | "(" <ausdruck> <operator> <ausdruck> ")" ) 
   //          	  [<space>]
   const int NO_POS=ausdruck.length();
   Bruch bruch1,
         bruch2,
	 ergebnis;
   char op;

   skipspace(ausdruck,pos);
   // pos steht auf erstem Nicht-Leerzeichen des Ausdrucks
   // oder bei String aus Leerzeichen auf NO_POS

   if (pos==NO_POS) {
      ergebnis.zaehler=0;
      ergebnis.nenner=0;
      ergebnis.vz=plus;
   }
   else {
      if (ausdruck.at(pos)=='+' || ausdruck.at(pos)=='-')
         // erwartet vollst. <bruch>; holen, berechnen
         ergebnis = bruchWert(ausdruck, pos);
      else
         if (ausdruck.at(pos)=='(') {
            // erwartet vollst. <klammerausdruck>; holen, berechnen
            ++pos; // von '(' auf Beginn des naechsten Ausdrucks setzen
	    bruch1=ausdruckWert(ausdruck, pos);
	    op=ausdruck.at(pos);
	    pos++; // von Operator auf Beginn des naechsten Ausdrucks setzen
	    bruch2=ausdruckWert(ausdruck,pos);
	    switch (op) {
               case '+': ergebnis=bruch1+bruch2; break;
               case '-': ergebnis=bruch1-bruch2; break;
               case '*': ergebnis=bruch1*bruch2; break;
               case '/': ergebnis=bruch1/bruch2; break;
               default:  cout << "Fehler im Ausdruck. Operator "
	                      << op << " nicht in +-*/\n";
	                 ergebnis.zaehler=0; 
                         ergebnis.nenner=0;
                         ergebnis.vz=plus;
	                 break;
            };
            ++pos; // von ')' auf naechstes Zeichen
         }
         else {
            cout << "Fehler im Ausdruck an Pos. " << pos << " \"(\" erwartet\n";
	    ergebnis.zaehler=0;
            ergebnis.nenner=0;
            ergebnis.vz=plus;
         }
   }
   skipspace(ausdruck,pos);
   return ergebnis;
}


// VORBEDINGUNG: ausdruck.at(pos) ist Beginn 
//               eines korrekten Bruches mit Vorzeichen
Bruch bruchWert(string ausdruck, int &pos) {
   Bruch ergebnis;
   int const NO_POS=ausdruck.length();
   // hole <bruch> in ausdruck ab Position pos
   // setze pos auf das erste Zeichen hinter <bruch>
   // gebe Bruch zurck

   // Vorzeichen
   if (ausdruck.at(pos)=='+')
      ergebnis.vz=plus;
   else
      ergebnis.vz=minus;
   ++pos; // von Vorzeichen auf erste Ziffer

   // Zaehler
   ergebnis.zaehler=zahlWert(ausdruck,pos);

   // Nenner
   if (pos==NO_POS) 
      // ausdruck endete mit ergebnis.zaehler
      // verhindert Absturz durch Pruefung von ausdruck.at(NO_POS)
      ergebnis.nenner=1;
   else
      // wenn ein Nenner folgt (ausdruck endet nie mit '/'!)
      if (ausdruck.at(pos)=='/' && istZiffer(ausdruck.at(pos+1))) {
         ++pos; // von '/' auf erste Ziffer des Nenners
         ergebnis.nenner=zahlWert(ausdruck,pos);
      }
      // wenn kein Nenner folgt
      else
         ergebnis.nenner=1;

   return ergebnis;
}

void skipspace(string ausdruck, int &pos) {
   // so lange ++pos, bis das aktuelle Zeichen kein Leerzeichen mehr ist
   // pos==NO_POS: String zu Ende
   const int NO_POS=ausdruck.length();
   if (pos != NO_POS) { // String Ende -- keine Leerzeichen mehr
                        // vermeidet die Auswertung ausdruck.at(NO_POS)
                        // was illegal ist und sonst einen Absturz bedeutet
      while (ausdruck.at(pos)==' ') {
         ++pos;
         if (pos == NO_POS) // StringEnde -- keine Leerzeichen mehr
            break;
      }
   }
}

bool istZiffer (char c) {
   return (c >= '0' && c <= '9');
}

int zahlWert (string zeichen, int &pos) {
   int f=0;
   const int NO_POS=zeichen.length();
   while (istZiffer(zeichen.at(pos))) {
      f= f*10 + (zeichen.at(pos) - '0');
      ++pos;
      if (pos==NO_POS) // Stringende erreicht
         break;
   }
   return f;
}

namespace BruchNS { 
   
Bruch operator+(Bruch bruch1, Bruch bruch2) {
   Bruch summenwert;
   gleichnamig(bruch1, bruch2);

   if (bruch1.vz == minus) bruch1.zaehler*=-1;
   if (bruch2.vz == minus) bruch2.zaehler*=-1;
   summenwert.zaehler = bruch1.zaehler + bruch2.zaehler;
   if (summenwert.zaehler < 0) {
      summenwert.vz = minus;
      summenwert.zaehler=abs(summenwert.zaehler);
   }
   else
      summenwert.vz = plus;
   summenwert.nenner  = bruch1.nenner;
   summenwert.kuerze();
   return summenwert;
}

Bruch operator-(Bruch bruch1, Bruch bruch2) {
   bruch2.zaehler *=-1;  // Subtraktion auf Addition zurueckfuehren
   return bruch1 + bruch2;
}

Bruch operator*(Bruch bruch1, Bruch bruch2) {
   Bruch mulwert;
   if (bruch1.vz == minus) bruch1.zaehler*=-1;
   if (bruch2.vz == minus) bruch2.zaehler*=-1;
   mulwert.zaehler = bruch1.zaehler * bruch2.zaehler;
   mulwert.nenner  = bruch1.nenner  * bruch2.nenner;
   if (mulwert.zaehler < 0) {
      mulwert.vz = minus;
      mulwert.zaehler = abs(mulwert.zaehler);
   }
   else
      mulwert.vz = plus;
   mulwert.kuerze();
   return mulwert;
}

Bruch operator/(Bruch bruch1, Bruch bruch2) {
   // Division auf Multiplikation mit den Kehrwert zurueckfuehren
   bruch2.kehrwert();
   return bruch1*bruch2;
}


void gleichnamig(Bruch &bruch1, Bruch &bruch2) {
   // gleichnamig machen, indem beide Brueche so erweitert 
   // werden, dass ihre Nenner das kgV beider Nenner sind.
   unsigned int kgv_nenner = kgv(bruch1.nenner,bruch2.nenner);
   bruch1.zaehler *= kgv_nenner / bruch1.nenner;
   bruch1.nenner   = kgv_nenner;
 
   bruch2.zaehler *= kgv_nenner / bruch2.nenner;
   bruch2.nenner   = kgv_nenner;
}

int kgv(int a, int b) {
      // kgV ist a multipliziert mit den Primfaktoren, die in b, aber nicht in a vorkommen
      //  "   "  b      "        "    "       "         "   " a    "   "     " b     "
      // diese ergeben sich aus b/<gemeinsame_Primfaktoren(a,b)>, d.i. b/ggT(a,b)
      return a * b/ggt(a,b);
}

int ggt(int a, int b) {
      // rekursive Berechnung des ggT ber Heron-Verfahren:
      //          { a          fuer a==b
      // ggT(a,b)={ ggT(a,b-a) fuer b>a
      //          { ggT(a-b,a) fuer b<a
      if (a==b)
         // Rekursionsende
         return a;
      else
         if (a>b)
            return ggt(a-b,b);
         else
            return ggt(a,b-a);
}

void Bruch::erweitere(int faktor) {
   if (faktor<0)
      if (vz==plus) vz=minus;
      else          vz=plus;
   faktor=abs(faktor);
   zaehler*=faktor;
   nenner *=faktor;
}
 
void Bruch::kuerze() {
   // da ggT(0,bruch.nenner) nicht bestimmbar:  
   if (zaehler == 0)
      return;
   unsigned int ggt_bruch = ggt(zaehler,nenner);
   zaehler = zaehler / ggt_bruch;
   nenner = nenner  / ggt_bruch;
}

void Bruch::lese() {
   cout << "Zhler eingeben: ";
   cin >> zaehler;
   cout << "Nenner eingeben: ";
   cin >> nenner;
   cout << endl;
   if (zaehler*nenner > 0)  // beide positiv oder beide negativ
      vz = plus;
   else 
      vz = minus;
   zaehler = abs(zaehler);
   nenner = abs(nenner);
}

void Bruch::schreibe() {
   cout.width(7);
   cout << zaehler << endl;
   if (vz == minus)   cout << "- -----\n";
   else               cout << "+ -----\n";
   cout.width(7);
   cout << nenner << endl << endl;
}

void Bruch::kehrwert() {
   int hilf; // Hilfsvariable zum Dreieckstausch
   hilf = nenner; 
   nenner = zaehler;
   zaehler = hilf;
}   

} // Ende namespace BruchNS

