/*
Programm: Aufg.Bruchrechnung
Zweck:    Berechnet Summe, Differenz, Produkt und Quotient von zwei 
          Bruechen und die gekuerzten Brueche
Aufgabe:  Entspricht dem Punkt "7.10.2 Aufgabe 2" im 
          Skript "Programmieren 1" von Prof. Letschert, in der 
          Version vom 2001-09-10
Autor:    Mattias Ansorg <ansis@gmx.de>
Lizenz:   public domain
Datum:    2001-12-06
*/

#include<iostream>
using namespace std;

namespace BruchNS { 
   enum Vorzeichen {plus, minus}; 
   struct Bruch { 
      Vorzeichen vz; 
      int zaehler; 
      int nenner; 
      void lese(); 
      void schreibe(); 
      void kuerze(); // soweit wie moeglich kuerzen 
      void erweitere(int); 
      void kehrwert();
   }; 
   Bruch add (Bruch, Bruch); 
   Bruch sub (Bruch, Bruch); 
   Bruch mult (Bruch, Bruch); 
   Bruch div (Bruch, Bruch); 
   void gleichnamig(Bruch &, Bruch &);
   int kgv(int, int);
   int ggt(int, int);   
} 

int main () { 
   using namespace BruchNS;
   Bruch b1,
         b2,
	 res;
	 
   // zwei Brueche einlesen 
   cout << "Eingabe von zwei Bruechen: \n";
   b1.lese();
   b2.lese();

   // Die Bruche in maximal gekuerzter Form ausgeben:
   cout << "Die Brche in maximal gekrzter Form: \n";
   b1.kuerze();
   b2.kuerze();
   b1.schreibe();
   b2.schreibe();

   // Summe der Brueche ausgeben
   cout << "Die Summe beider Brche: \n";
   res = add(b1,b2);
   res.schreibe();
   
   // Differenz der Brueche ausgeben
   cout << "Die Differenz beider Brche: \n";
   res = sub(b1,b2);
   res.schreibe();
	   
   // Produkt der Brueche ausgeben
   cout << "Das Produkt beider Brche: \n";
   res = mult(b1,b2);
   res.schreibe();
   
   // Quotient der Brueche ausgeben
   cout << "Der Quotient beider Brche: \n";
   res = div(b1,b2);
   res.schreibe();
} 

namespace BruchNS { 
   
Bruch add(Bruch bruch1, Bruch bruch2) {
   Bruch summenwert;
   gleichnamig(bruch1, bruch2);

   if (bruch1.vz == minus) bruch1.zaehler*=-1;
   if (bruch2.vz == minus) bruch2.zaehler*=-1;
   summenwert.zaehler = bruch1.zaehler + bruch2.zaehler;
   if (summenwert.zaehler < 0) {
      summenwert.vz = minus;
      summenwert.zaehler=abs(summenwert.zaehler);
   }
   else
      summenwert.vz = plus;
   summenwert.nenner  = bruch1.nenner;
   summenwert.kuerze();
   return summenwert;
}

Bruch sub(Bruch bruch1, Bruch bruch2) {
   bruch2.zaehler *=-1;  // Subtraktion auf Addition zurueckfuehren
   return add(bruch1, bruch2);
}

Bruch mult(Bruch bruch1, Bruch bruch2) {
   Bruch mulwert;
   if (bruch1.vz == minus) bruch1.zaehler*=-1;
   if (bruch2.vz == minus) bruch2.zaehler*=-1;
   mulwert.zaehler = bruch1.zaehler * bruch2.zaehler;
   mulwert.nenner  = bruch1.nenner  * bruch2.nenner;
   if (mulwert.zaehler < 0) {
      mulwert.vz = minus;
      mulwert.zaehler = abs(mulwert.zaehler);
   }
   else
      mulwert.vz = plus;
   mulwert.kuerze();
   return mulwert;
}

Bruch div(Bruch bruch1, Bruch bruch2) {
   // Division auf Multiplikation mit den Kehrwert zurueckfuehren
   bruch2.kehrwert();
   return mult(bruch1, bruch2);
}


void gleichnamig(Bruch &bruch1, Bruch &bruch2) {
   // gleichnamig machen, indem beide Brueche so erweitert 
   // werden, dass ihre Nenner das kgV beider Nenner sind.
   unsigned int kgv_nenner = kgv(bruch1.nenner,bruch2.nenner);
   bruch1.zaehler *= kgv_nenner / bruch1.nenner;
   bruch1.nenner   = kgv_nenner;
 
   bruch2.zaehler *= kgv_nenner / bruch2.nenner;
   bruch2.nenner   = kgv_nenner;
}

int kgv(int a, int b) {
      // kgV ist a multipliziert mit den Primfaktoren, die in b, aber nicht in a vorkommen
      //  "   "  b      "        "    "       "         "   " a    "   "     " b     "
      // diese ergeben sich aus b/<gemeinsame_Primfaktoren(a,b)>, d.i. b/ggT(a,b)
      return a * b/ggt(a,b);
}

int ggt(int a, int b) {
      // rekursive Berechnung des ggT ber Heron-Verfahren:
      //          { a          fuer a==b
      // ggT(a,b)={ ggT(a,b-a) fuer b>a
      //          { ggT(a-b,a) fuer b<a
      if (a==b)
         // Rekursionsende
         return a;
      else
         if (a>b)
            return ggt(a-b,b);
         else
            return ggt(a,b-a);
}

void Bruch::erweitere(int faktor) {
   if (faktor<0)
      if (vz==plus) vz=minus;
      else          vz=plus;
   faktor=abs(faktor);
   zaehler*=faktor;
   nenner *=faktor;
}
 
void Bruch::kuerze() {
   // da ggT(0,bruch.nenner) nicht bestimmbar:  
   if (zaehler == 0)
      return;
   unsigned int ggt_bruch = ggt(zaehler,nenner);
   zaehler = zaehler / ggt_bruch;
   nenner = nenner  / ggt_bruch;
}

void Bruch::lese() {
   cout << "Zhler eingeben: ";
   cin >> zaehler;
   cout << "Nenner eingeben: ";
   cin >> nenner;
   cout << endl;
   if (zaehler*nenner > 0)  // beide positiv oder beide negativ
      vz = plus;
   else 
      vz = minus;
   zaehler = abs(zaehler);
   nenner = abs(nenner);
}

void Bruch::schreibe() {
   cout.width(7);
   cout << zaehler << endl;
   if (vz == minus)   cout << "- -----\n";
   else               cout << "+ -----\n";
   cout.width(7);
   cout << nenner << endl << endl;
}

void Bruch::kehrwert() {
   int hilf; // Hilfsvariable zum Dreieckstausch
   hilf = nenner; 
   nenner = zaehler;
   zaehler = hilf;
}   

}
